from django.apps import AppConfig


class CouncilmaticConfig(AppConfig):
    name = 'councilmatic_core'
    verbose_name = "Councilmatic"

    def ready(self):
        import councilmatic_core.signals.handlers  # noqa



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-01-08 16:47
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0034_event_guid'),
    ]

    operations = [
        migrations.AddField(
            model_name='bill',
            name='html_text',
            field=models.TextField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.1.9 on 2019-06-26 21:17

import django.contrib.gis.db.models.fields
import django.core.files.storage
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0004_auto_20171005_2028'),
        ('councilmatic_core', '0047_update_filepath'),
    ]

    operations = [
        migrations.DeleteModel(
            name='Post',
        ),
        migrations.CreateModel(
            name='Post',
            fields=[
                ('post', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='councilmatic_post', serialize=False, to='core.Post')),
                ('shape', django.contrib.gis.db.models.fields.GeometryField(null=True, srid=4326)),
            ],
            options={
                'abstract': False,
            },
            bases=('core.post',),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-06-27 19:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0041_event_extras'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventagendaitem',
            name='plain_text',
            field=models.TextField(null=True),
        ),
    ]



import os

import django
from django.conf import settings
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.db import connection


class Command(BaseCommand):
    help = 'Refresh the test fixtures from an instance database'

    def _id_gen(self, table):
        with connection.cursor() as cursor:
            cursor.execute('SELECT id FROM opencivicdata_{} LIMIT 3'.format(table))
            ids = [row[0] for row in cursor]
        return ids

    def handle(self, *args, **options):
        current_directory = os.path.dirname(__file__)

        outfiles = []

        for table in ('bill', 'person', 'organization', 'event'):
            args = ['dump_object', 'councilmatic_core.{}'.format(table.title())] + self._id_gen(table) + ['--kitchensink']
            outfile = os.path.join(current_directory, '..', '..', '..', '{}.json'.format(table))
            self.stdout.write('writing {}'.format(outfile))
            with open(outfile, 'w') as f:
                call_command(*args, stdout=f)
            outfiles.append(outfile)

        self.stdout.write('merging outfiles')
        
        with open(os.path.join(current_directory, '..', '..', '..', 'tests', 'fixtures', 'test_data.json'), 'w') as f:
            call_command('merge_fixtures', *outfiles, stdout=f)

        for f in outfiles:
            self.stdout.write('removing {}'.format(f))
            os.remove(f)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-04-03 14:03
from __future__ import unicode_literals
import re

from django.conf import settings
from django.db import migrations
from django.db import connection, transaction
from django.utils.text import slugify

class Migration(migrations.Migration):
    def unmangle_identifier(apps, schema_editor):
        '''
        The `fix_bill_id` function mangled nearly all Chicago bill identifiers by adding a space, e.g., 'Or2018-1245' (original) becomes 'Or 2018-1245' (mangled).

        
        This migration unmangles identifiers and changes the slugs accordingly.
        '''

        Bill = apps.get_model('councilmatic_core', 'Bill')

        # Chicago
        if settings.OCD_CITY_COUNCIL_NAME == 'Chicago City Council':
            added_space = r'^([A-Za-z]+)\s([-\d]+)$'

            with transaction.atomic(), connection.cursor() as cursor:
                cursor.execute("""
                    DECLARE bills_cursor CURSOR FOR
                    SELECT identifier FROM councilmatic_core_bill as bill
                    WHERE bill.identifier ~* '(^([A-Za-z]+)\s([-\d]+)$)'
                """)

                while True:
                    cursor.execute("FETCH 4000 FROM bills_cursor")
                    bill_chunk = cursor.fetchall()

                    if not bill_chunk:
                        break
                    else: 
                        for bill in bill_chunk:
                            identifier = bill[0]
                            match = re.match(added_space, identifier)
                            unmangled_identifier = '{mangled_prefix}{count}'.format(mangled_prefix=match.group(1), 
                                                                                    count=match.group(2))

                            print('{} becomes {}'.format(identifier, unmangled_identifier))
                            try:
                                duplicate = Bill.objects.get(identifier=unmangled_identifier)
                                print('{} - duplicate found. Deleting.'.format(unmangled_identifier))
                                duplicate.delete()
                            except Bill.DoesNotExist: 
                                pass

                            new_slug = slugify(unmangled_identifier)
                            Bill.objects.filter(identifier=identifier).update(identifier=unmangled_identifier, slug=new_slug)

    dependencies = [
        ('councilmatic_core', '0037_auto_20180326_1631'),
    ]

    operations = [
        migrations.RunPython(unmangle_identifier)
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-04-12 15:53
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0038_auto_20180403_1003'),
    ]

    operations = [
        migrations.CreateModel(
            name='EventMedia',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.CharField(max_length=555)),
            ],
        ),
        migrations.AddField(
            model_name='eventmedia',
            name='event',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='media_urls', to='councilmatic_core.Event'),
        ),
        migrations.RunSQL('''
            INSERT INTO councilmatic_core_eventmedia (event_id, url)
              SELECT
                ocd_id,
                media_url
              FROM councilmatic_core_event AS event
              WHERE media_url IS NOT NULL
        '''),
        migrations.RemoveField(
            model_name='event',
            name='media_url',
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0002_auto_20151210_1342'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bill',
            name='ocd_updated_at',
            field=models.DateTimeField(default=None),
        ),
        migrations.AlterField(
            model_name='person',
            name='ocd_id',
            field=models.CharField(unique=True, max_length=100, null=True),
        ),
        migrations.AlterField(
            model_name='person',
            name='slug',
            field=models.CharField(unique=True, max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-04-12 16:15
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
from django.utils.timezone import utc


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0039_multi_event_media'),
    ]

    operations = [
        migrations.AddField(
            model_name='eventmedia',
            name='note',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='eventmedia',
            name='updated_at',
            field=models.DateTimeField(auto_now=True, default=datetime.datetime(2018, 4, 12, 16, 15, 50, 262081, tzinfo=utc)),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-10-12 17:04
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0010_auto_20160120_1248'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='organization',
            name='id',
        ),
        migrations.RemoveField(
            model_name='post',
            name='id',
        ),
        migrations.AlterField(
            model_name='organization',
            name='ocd_id',
            field=models.CharField(max_length=100, primary_key=True, serialize=False, unique=True),
        ),
        migrations.AlterField(
            model_name='post',
            name='ocd_id',
            field=models.CharField(max_length=100, primary_key=True, serialize=False, unique=True),
        ),
    ]



import os
import logging
import logging.config
import sqlalchemy as sa
import requests
import tempfile
import itertools

from django.core.management.base import BaseCommand
from django.conf import settings
from django.db.models import Max, Q

from opencivicdata.legislative.models import BillDocumentLink
from councilmatic_core.models import BillDocument

logging.config.dictConfig(settings.LOGGING)
logging.getLogger("requests").setLevel(logging.WARNING)
logger = logging.getLogger(__name__)

DB_CONN = 'postgresql://{USER}:{PASSWORD}@{HOST}:{PORT}/{NAME}'

engine = sa.create_engine(DB_CONN.format(**settings.DATABASES['default']),
                          convert_unicode=True,
                          server_side_cursors=True)

class Command(BaseCommand):
    help = 'Converts bill attachments into plain text'

    def add_arguments(self, parser):
        parser.add_argument(
            '--update_all',
            default=False,
            action='store_true',
            help='Add or update plain text for all bill attachments.')

    def handle(self, *args, **options):
        self.update_all = options['update_all']
        self.add_plain_text()

    def get_document_url(self):
        # Only apply this query to most recently updated (or created) bill documents.
        max_updated = BillDocument.objects.all().aggregate(max_updated_at=Max('bill__updated_at'))['max_updated_at']

        is_null = Q(document__councilmatic_document__full_text__isnull=True)
        is_file = Q(url__iendswith='pdf') | Q(url__iendswith='docx') | Q(url__iendswith='docx')
        after_max_update = Q(document__bill__updated_at__gt=max_updated)

        if max_updated is None or self.update_all:
            qs = BillDocumentLink.objects.filter(is_null & is_file)
        else:
            qs = BillDocumentLink.objects.filter(is_null & is_file & after_max_update)

        for item in qs:
            yield item.url, item.document.id

    def convert_document_to_plaintext(self):
        # textract is a heavy dependency. In order to test this code without
        # installing it, import the library here.
        import textract

        for document_data in self.get_document_url():
            document_data = dict(document_data)
            url = document_data['url']
            document_id = document_data['id']
            response = requests.get(url)
            # Sometimes, Metro Legistar has a URL that retuns a bad status code (e.g., 404 from http://metro.legistar1.com/metro/attachments/95d5007e-720b-4cdd-9494-c800392b9265.pdf).
            # Skip these documents.
            if response.status_code != 200:
                logger.error('Document URL {} returns {} - Could not get attachment text!'.format(url, response.status_code))
                continue

            extension = os.path.splitext(url)[1]

            with tempfile.NamedTemporaryFile(suffix=extension) as tfp:
                tfp.write(response.content)

                try:
                    plain_text = textract.process(tfp.name)
                except textract.exceptions.ShellError as e:
                    logger.error('{} - Could not convert Councilmatic Document ID {}!'.format(e, document_id))
                    continue

                logger.info('Councilmatic Document ID {} - conversion complete'.format(document_id))

            yield {'plain_text': plain_text.decode('utf-8'), 'id': document_id}

    def add_plain_text(self):
        '''
        Metro has over 2,000 attachments that should be converted into plain text.
        When updating all documents with `--update_all`, this function insures that the database updates only 20 documents per connection (mainly, to avoid unexpected memory consumption).
        It fetches up to 20 elements from a generator object, runs the UPDATE query, and then fetches up to 20 more.

        Inspired by: https://stackoverflow.com/questions/30510593/how-can-i-use-server-side-cursors-with-django-and-psycopg2/41088159#41088159

        More often, this script updates just a handful of documents: so, the incremental, fetch-just-20 approach may prove unnecessary. Possible refactor?
        '''
        update_statement = '''
            UPDATE councilmatic_core_billdocument AS bill_docs
            SET full_text = :plain_text
            WHERE bill_docs.document_id = :id
        '''

        plaintexts = self.convert_document_to_plaintext()

        while True:
            logger.info('Updating documents with plain text...')
            plaintexts_fetched_from_generator = list(itertools.islice(plaintexts, 20))

            if not plaintexts_fetched_from_generator:
                break
            else:
                with engine.begin() as connection:
                    connection.execute(sa.text(update_statement), plaintexts_fetched_from_generator)

        logger.info('SUCCESS')



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-10-05 19:37
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0042_eventagendaitem_plain_text'),
    ]

    operations = [
        migrations.AddField(
            model_name='membership',
            name='extras',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-councilmatic',
    version='2.5.0',
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    license='MIT License',  # example license
    description='Core functions for councilmatic.org family',
    long_description=README,
    url='http://councilmatic.org/',
    author='DataMade, LLC',
    author_email='info@datamade.us',
    install_requires=['requests>=2.20,<2.21',
                      'opencivicdata>=2.2.1<2.3',
                      'pytz>=2015.4',
                      'django-haystack>=2.8.0,<2.9',
                      'Django>=2.0,<2.2',
                      'django-proxy-overrides>=0.2.1',
                      'python-dateutil>=2.7<2.8',
                      'pysolr>=3.8,<3.9',
                      'psycopg2-binary>=2.7<2.8',
                      'django-adv-cache-tag==1.1.2',
                      'boto==2.38.0',
                      'sqlalchemy'],
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',  # example license
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        # Replace these appropriately if you are stuck on Python 2.
        'Programming Language :: Python :: 3.4',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-03-31 15:18
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0027_merge'),
    ]

    operations = [
        migrations.AddField(
            model_name='event',
            name='media_url',
            field=models.CharField(default=None, max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2018-03-02 17:47
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('councilmatic_core', '0035_bill_html_text'),
    ]

    operations = [
        migrations.CreateModel(
            name='Jurisdiction',
            fields=[
                ('ocd_id', models.CharField(max_length=100, primary_key=True, serialize=False, unique=True)),
                ('name', models.CharField(max_length=300)),
                ('classification', models.CharField(max_length=50)),
                ('url', models.CharField(max_length=2000)),
            ],
        ),
        migrations.AddField(
            model_name='organization',
            name='jurisdiction',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='organizations', to='councilmatic_core.Jurisdiction'),
        ),
    ]



import json

from django.core.management.base import BaseCommand, CommandError
from django.contrib.gis.geos import GEOSGeometry

from councilmatic_core import models


class Command(BaseCommand):
    help = "Import boundary shapefiles for Post entities"

    def add_arguments(self, parser):
        parser.add_argument(
            'geojson_file',
            help=(
                'The location of the GeoJSON file containing shapes for each Division, '
                'relative to the project root. The file should be formatted as a '
                'GeoJSON FeatureCollection where each Feature A) corresponds to a distinct '
                'Division and B) has a "division_id" attribute in the "properties" object. '
            )
        )

    def handle(self, *args, **options):
        self.stdout.write('Populating shapes for Posts...')
        shapes_populated = 0

        with open(options['geojson_file']) as shapef:
            shapes = json.load(shapef)

        features = self._get_or_raise(
            shapes,
            'features',
            'Could not find the "features" array in the input file.'
        )

        for feature in features:
            shape = self._get_or_raise(
                feature,
                'geometry',
                'Could not find a "geometry" key in the Feature.'
            )
            properties = self._get_or_raise(
                feature,
                'properties',
                'Could not find a "properties" key in the Feature.'
            )
            division_id = self._get_or_raise(
                properties,
                'division_id',
                'Could not find a "division_id" key in the Feature properties.'
            )

            models.Post.objects.filter(division_id=division_id).update(
                shape=GEOSGeometry(json.dumps(shape))
            )
            shapes_populated += 1

        self.stdout.write(
            self.style.SUCCESS(
                'Populated {} shapes'.format(str(shapes_populated))
            )
        )

    def _get_or_raise(self, dct, key, msg):
        """
        Check to see if 'dct' has a key corresponding to 'key', and raise an
        error if it doesn't.
        """
        format_prompt = (
            'Is the input file formatted as a GeoJSON FeatureCollection '
            'where each feature has a "division_id" property?'
        )
        if not dct.get(key):
            raise CommandError(msg + ' ' + format_prompt)
        else:
            return dct[key]
